//*********************************************************************************************************
//   Program written for Arduino Nano to drive Vishay "AGC 0" type presence sensors consisting of e.g. TSSP4056
//   or TSSP4038 (connected to pin D8) and VSLB3940 IR emitter (connected with a series resistor to pin D3).
//   The user can vary the burst frequency to adjust the system gain, the burst repetition time to adjust the power
//   consumption, and the number of cycles in a burst. The onboard LED is used to indicate a positive sense event.
//   I/O is via the serial monitor used at 9600 baud and set to send Newline characters.
//   Author: John Fisher
//*********************************************************************************************************

const int BUFFER_SIZE = 6;                                      // buffer for serial input, max 6 digits for burstrep_count
volatile uint8_t BurstCount = 0;                                // global variable for number of carrier cycles in burst
uint8_t Pause = 10;                                             // burst repetition time in miliseconds
volatile uint16_t InputCapture = 0;                             // used to capture the sensor pulse-length from 16 bit timer1
enum {INACTIVE, ACTIVE} State = INACTIVE;                       // INACTIVE is waiting for serial inputs, ACTIVE is sending bursts
volatile enum {rx_IDLE, rx_PENDING, rx_DONE} rx_stat = rx_IDLE; // rx_IDLE = no pulse yet seen, rx_PENDING= Rx is capturing a pulse, rx_DONE = Rx finished capturing the pulse
volatile enum {tx_PENDING, tx_DONE} tx_stat = tx_DONE;          // tx_PENDING = tx is sending a burst, tx_DONE = tx finished sending a burst
uint8_t HalfPeriod = 143;                                       // (Period_56kHz/2) / Period_16MHz

//*********************************************************************************************************
//   Function to input a line of alpha numeric characters. Requires an array and its size.
//   Returns the number of characters read.
//*********************************************************************************************************
int getline(char* buf, int bufsize){
  for (int bufctr = 0; bufctr < bufsize; bufctr++) {
    while (Serial.available() == 0) { }               // do nothing if no chars received yet
    char onechar= Serial.read();                      // read one character
    if (onechar== '\n') {                             // if EOL
      buf[bufctr] = 0;                                // then add string terminator
      return bufctr;                                  // and return length of string
    }                                                 // end if
    else {
      buf[bufctr] = onechar;                          // else add character to buf
    }                                                 // end else
  }                                                   // end for

  // If no EOL was received, issue warning
  Serial.println("Warning, either too many characters were sent, or");
  Serial.println("the Serial Monitor is not set to send Newline characters.");
  Serial.println("Please correct, and restart the monitor.");
  buf[0] = 0;                                         // discard the input
  return -1;                                          // return error
}                                                     // end getline

//*********************************************************************************************************
//  Interrupt service routine
//  Timer2 overflow interrupt used to count carrier cycles
//*********************************************************************************************************
ISR(TIMER2_OVF_vect) {
  static uint16_t carrier_count = 0;       // persistent counter used to count carrier cycles
  carrier_count++;                         // increment carrier cycle counter
  if (carrier_count >= BurstCount) {       // if burst is completed
    carrier_count = 0;                     // reset counter for start of new cycle
    TCCR2B &= 0B11111000;                  // clear the CS bits, which halts Timer2
    TCCR2A = 0;                            // return pin D3 to normal IO mode for now
    PORTD &= ~(1 << DDD3);                 // turn off emitter when idle (until next burst)
    tx_stat = tx_DONE;                     // change tx status to finished sending a burst
  }                                        // end if
}                                          // end ISR

//*********************************************************************************************************
//  Interrupt service routine
//  Timer1 edge based interrupts to capture the sensor pulse
//*********************************************************************************************************
ISR(TIMER1_CAPT_vect) {
  uint8_t scratch = TCCR1B & (1 << ICES1); // result should be 64 if set to trigger on rising edge, 0 if set to trigger on falling edge
  switch (scratch) {                       // switch depending on the ICES1 bit
    case 0:                                // ICES1 = 0, have falling edge trigger
      TCNT1 = 0;                           // reset Timer1
      TCCR1B |= (1 << ICES1);              // set interrupt to rising edge trigger
      rx_stat = rx_PENDING;                // set status to the Rx is capturing a pulse
      break;
    case 64:                               // ICES1 = 1, have rising edge trigger
      InputCapture = ICR1;                 // save the capture register value
      TCCR1B &= ~(1 << ICES1);             // set interrupt to falling edge trigger
      rx_stat = rx_DONE;                   // set status to the Rx has finished capturing a pulse
      break;
  }                                        // end switch
}                                          // end ISR


//**************************************************************************************************************************
//  Routine to send a single pulse. Function arguments: half-period of the carrier, used in Timer2 = (1/2xburstfreq) / (1/clkfreq), the
//  duty cycle of the carrier in %, and the number of cycles in a burst. The function returns the width of the sensor pulse from Timer1,
//  or -1 if no reflection is seen, or 2 if an invalid reflection is seen, i.e. outside the receiver spec.
//****************************************************************************************************************************
int16_t SendBurst (uint8_t HalfPeriod = 143, uint8_t BurstLength = 25, uint8_t Duty = 50) {

  int16_t Timer1Count;                        // holds the return value
  
  // -------------------------------------  set up timer 2 for transmitter --------------
  BurstCount = BurstLength;                   // set global variable visible to Timer2 ISR
  TCCR2A = (1 << WGM20) | (1 << COM2B1);      // clear OC2B on match counting up and set counting down, Mode5: PWM phase correct with TOP=OCR2A
  TCCR2B = (1 << WGM22);                      // FOC not used, no pre-scaler, final bit of Mode5 setting, cs bits are all zero so timer is off
  TIMSK2 = (1 << TOIE2);                      // enable timer2 overflow interrupt
  OCR2A  = HalfPeriod;                        // set timer2 TOP value
  OCR2B  = HalfPeriod * Duty / 100;           // set switching value for the output pin
  TCNT2 = 0;                                  // reset Timer2 to zero

  // -------------------------------------  set up timer 1 for receiver -----------------
  uint16_t PulsTolHi = (BurstLength + 6) * HalfPeriod / 32; // output pulse upper tolerance, see datasheet fig. 1
  uint16_t PulsTolLo = (BurstLength - 5) * HalfPeriod / 32; // output pulse lower tolerance, wider than datasheet to account for off BP frequencies
  TCCR1A = 0;                                 // normal mode, no output pins connected, TOP=0xFFFF
  rx_stat = rx_IDLE;                          // set the receiver state to idle
  TCCR1B = (1 << ICNC1) | (0 << ICES1) | (1 << CS11) | (1 << CS10); // noise cancelling input on, input capture edge select to falling edge, prescaler 64
  TIMSK1 = (1 << ICIE1);                      // enable timer1 input capture interrupt (transfers TCNT1 to ICR1 upon the edge defined)

  // -------------------------------------  end set up, start the transmission -----------------------------
  tx_stat = tx_PENDING;                       // tx status is sending burst
  TIFR1 |= (1 << ICF1);                       // make sure any pending interrupt is cleared...potential noise or interference
  TIFR2 |= (1 << TOV2);                       // make sure any pending interrupt is cleared
  sei();                                      // allow interrupts
  TCCR2B |= (1 << CS20);                      // start Timer2 with no pre-scale
  while (tx_stat != tx_DONE )                 // if the transmitted burst has not been completely sent
    ;                                         // normally do nothing until tx_stat = tx_DONE
  switch (rx_stat) {                          // check status of receiver
    case rx_IDLE:                             // if receiver is idle, no reflection was seen at the highest sensitivity, assume that object is out of range
      Timer1Count = -1;                       // -1 means no reflection
      break;
    case rx_PENDING:                          // if receiver is pending, we saw a reflection and we are currently measuring it
      while (rx_stat == rx_PENDING)           // can do nothing useful so wait for it to end
        ;                                     // do nothing until rx_stat = rx_DONE
    //break;                                  // no break here! once rx_stat != rx_PENDING, we want to fall through to the next case, not exit the switch!!
    case rx_DONE:                             // if receiver is done, the reflected pulse was measured, now check its validity
      Timer1Count = InputCapture;             // store the timer1 value
      rx_stat = rx_IDLE;                      // rx finished, set status to idle
      if ( !((Timer1Count > PulsTolLo) && (Timer1Count < PulsTolHi))) { // if we received an out of spec pulse
        Timer1Count = 2;                      // return magic number 2 to signal out of spec pulse
      }
      break;                                  // this is the end of the case: rx_stat = rx_IDLE
  }

  return Timer1Count;
}                                             // end SendBurst

//***********************************************************************************************************
//  Arduino setup()
//***********************************************************************************************************

#define BAUD_RATE 9600

void setup() {

  Serial.begin(BAUD_RATE);                                    // initialize serial baud rate

  DDRD |= (1 << DDD3);                                        // set emitter pin D3 as an output
  PORTD &= ~(1 << DDD3);                                      // turn off emitter pin D3 until needed later
  DDRB &= ~(1 << DDB0);                                       // define Arduino D8, AVR PB0 as input
  if ((PORTB & (1 << PORTB0)) == 0)   PORTB |= (1 << PORTB0); // this activates the internal pull-up resistor in case the IR sensor is disconnected

}                                                             // end setup

//***********************************************************************************************************
//  Arduino loop()
//***********************************************************************************************************

void loop() {

  while (true) {                                       // avoids extra processing at end of every loop

    if (BurstCount == 0){                              // if BurstCount variable invalid
      char line[BUFFER_SIZE];                          // temporary input buffer
      Serial.println("*PLEASE ENTER THE SIGNAL SETTINGS*");
      uint8_t Freq = 0;
 
      do {
        line[0] = 0;                                   // reset buffer
        Serial.println("Carrier Frequency in kHz (36 - 56), highest sensitivity at 56");
        getline(line, sizeof(line));                   // get carrier frequency
        Serial.println(line);                          // echo it
        Freq = atoi(line);                             // convert ASCII to integer
      } while (Freq < 36 || Freq > 56);                // check for bad input
        
      HalfPeriod = 8000/Freq;                          // (1/2xfreq)/(1/16MHz)
    
      do {
        line[0] = 0;                                   // reset buffer
        Serial.println("Burst Repetition in Milliseconds (5 - 200), lowest power at 200");
        getline(line, sizeof(line));                   // get burst repetition time in ms
        Serial.println(line);                          // echo it
        Pause = atoi(line);                            // convert ASCII to integer
      } while (Pause < 5 || Pause > 200);              // check for bad input

      do {
        line[0] = 0;                                   // reset buffer
        Serial.println("Burst Length in Cycles (10 - 50)");
        getline(line, sizeof(line));                   // get burst length in cycles
        Serial.println(line);                          // echo it
        BurstCount = atoi(line);                       // convert ASCII to integer
      } while (BurstCount < 10 || BurstCount > 50);    // check for bad input
    
      Serial.println("Press key '1' for start sending, key '0' for stop sending");
      Serial.println(" ");
      delay(2000);
    }                                                  // end if

//***********************************************************************************************************

    if (Serial.available()){                       // is input available?
      char line[BUFFER_SIZE];                      // temporary input buffer
      getline(line, sizeof(line));                 // get command
      char command = line[0];                      // get first char for the switch statement
      switch (command){
        case '1':                                  // start sending and receiving on "1"
        State = ACTIVE;
        Serial.println("SENDING");            
        break;
      default:                                     // stop sending and receiving on any other character
        State = INACTIVE;
        BurstCount = 0;                            // set variable invalid
        Serial.println("STOP SENDING");
        Serial.println(" ");
        break;
      }                                            // end switch
    }                                              // end if

//***********************************************************************************************************

    if (State == ACTIVE){

      int16_t pulsewidth = 0;                                     // variable to hold the return value from function SendBurst
      if ((pulsewidth = SendBurst(HalfPeriod, BurstCount)) > 2) { // if we see a valid reflection
        PORTB |= (1<<5);                                          // turn on LED or do something else useful here!
        delay(Pause);                                             // burst repetition time
        PORTB &= ~(1<<5);                                         // turn off LED
      }                                                           // end if
      else (delay(Pause));                                        // burst repetition time
    }                                                             // end if
    
  }                                                               // end while true
}                                                                 // end loop

